#------------------------------------------------------------------------------
#$Author: antanas $
#$Date: 2015-07-31 15:56:51 +0200 (Fri, 31 Jul 2015) $ 
#$Revision: 3775 $
#$URL: svn://www.crystallography.net/cod-tools/trunk/src/lib/perl5/COD/CIF/Data.pm $
#------------------------------------------------------------------------------
#*
#  Common subroutines for extracting data/creating data structures from
#  the parsed CIF data structures.
#**

package COD::CIF::Data;

use strict;
use warnings;
use COD::Spacegroups::Lookup::COD;
use COD::Spacegroups::Names;
use COD::UserMessage qw(warning error prefix_dataname);

require Exporter;
our @ISA = qw(Exporter);
our @EXPORT_OK = qw(
    get_cell
    get_content_encodings
    get_symmetry_operators
);

my %sg_name_abbrev =
    map { my $key = $_->[1]; $key =~ s/\s+//g; ( $key, $_->[2] ) }
    @COD::Spacegroups::Names::names,
    map { [ $_->{number}, $_->{hermann_mauguin}, $_->{universal_h_m} ] }
    @COD::Spacegroups::Lookup::COD::table,
    @COD::Spacegroups::Lookup::COD::extra_settings;

#===============================================================#
# Extract unit cell angles and lengths.

# Accepts
#     values - a hash where a data from the CIF file is stored
# Returns
#     cell_lengths_and_angles - an array  with stored information.

sub get_cell($$$@)
{
    my( $values, $filename, $dataname, $options ) = @_;
    $options = {} unless $options;

    $dataname = prefix_dataname($dataname);

    my @cell_lengths_and_angles;

    for my $cif_tag (qw(_cell_length_a
                        _cell_length_b
                        _cell_length_c
                        ))
    {
        if( exists $values->{$cif_tag} ) {
            push(@cell_lengths_and_angles, $values->{$cif_tag}[0]);
            $cell_lengths_and_angles[-1] =~ s/\(\d+\)$//;
        } elsif( $options->{silent} ) {
            push(@cell_lengths_and_angles, undef);
        } else {
            error( $0, $filename, $dataname,
                   "cell length '$cif_tag' not present", undef );
        }
    }

    for my $cif_tag (qw(_cell_angle_alpha
                        _cell_angle_beta
                        _cell_angle_gamma
                        ))
    {
        if( exists $values->{$cif_tag} ) {
            push( @cell_lengths_and_angles, $values->{$cif_tag}[0] );
            $cell_lengths_and_angles[-1] =~ s/\(\d+\)$//;
        } elsif( $options->{silent} ) {
            push(@cell_lengths_and_angles, undef);
        } else {
            warning( $0, $filename, $dataname,
                     "cell angle '$cif_tag' not present",
                     "taking default value 90 degrees" );
            push( @cell_lengths_and_angles, 90 );
        }
    }

    return @cell_lengths_and_angles;
}

sub get_symmetry_operators($$)
{
    my ( $dataset, $filename ) = @_;

    my $values = $dataset->{values};
    my $dataname = prefix_dataname($dataset->{name});

    my $sym_data;

    if( exists $values->{"_space_group_symop_operation_xyz"} ) {
        $sym_data = $values->{"_space_group_symop_operation_xyz"};
    } elsif( exists $values->{"_symmetry_equiv_pos_as_xyz"} ) {
        $sym_data = $values->{"_symmetry_equiv_pos_as_xyz"};
    }

    if( !defined $sym_data ) {
        for my $tag (qw( _space_group_name_Hall
                         _space_group_name_hall
                         _symmetry_space_group_name_Hall 
                         _symmetry_space_group_name_hall 
                     )) {
            if( exists $values->{$tag} ) {
                my $hall = $values->{$tag}[0];
                next if $hall eq '?';
                $sym_data = lookup_symops("hall", $hall);

                if( !$sym_data ) {
                    error( $0, $filename, $dataname,
                           "$tag value '$hall' is not recognised", undef );
                } else {
                    last
                }
            }
        }
    }

    if( !defined $sym_data ) {
        for my $tag (qw( _space_group_name_H-M_alt
                         _space_group_name_h-m_alt
                         _space_group.name_H-M_full
                         _space_group.name_h-m_full
                         _symmetry_space_group_name_H-M
                         _symmetry_space_group_name_h-m
                    )) {
            if( exists $values->{$tag} ) {
                my $h_m = $values->{$tag}[0];
                next if $h_m eq '?';
                $sym_data = lookup_symops("hermann_mauguin", $h_m);

                if( !$sym_data ) {
                    error( $0, $filename, $dataname,
                           "$tag value '$h_m' is not recognised", undef );
                } else {
                    last
                }
            }
        }
    }

    if( not defined $sym_data ) {
        for my $tag (qw( _space_group_ssg_name
                         _space_group_ssg_name_IT
                         _space_group_ssg_name_WJJ
                    )) {
            if( exists $values->{$tag} ) {
                my $ssg_name = $values->{$tag}[0];
                next if $ssg_name eq '?';

                my $h_m = $ssg_name;
                $h_m =~ s/\(.*//g;
                $h_m =~ s/[\s_]+//g;

                $sym_data = lookup_symops("hermann_mauguin", $h_m);

                if( !$sym_data ) {
                    error( $0, $filename, $dataname,
                           "$tag value '$ssg_name' yielded H-M " .
                           "symbol '$h_m' which is not in our tables", undef );
                } else {
                    last
                }
            }
        }
    }

    if( not defined $sym_data ) {
        error( $0, $filename, $dataname, 'neither symmetry operators, '
             . 'nor Hall spacegroup symbol, nor Hermann-Mauguin spacegroup '
             . 'symbol could be processed', undef );
        die;
    }

    return $sym_data;
}

sub get_content_encodings($$)
{
    my ( $dataset, $filename ) = @_;

    my $values = $dataset->{values};
    my $dataname = prefix_dataname($dataset->{name});

    if( !exists $values->{_tcod_content_encoding_id} ||
        !exists $values->{_tcod_content_encoding_layer_type} ) {
        return undef;
    }

    my %encodings;

    for( my $i = 0; $i < @{$values->{_tcod_content_encoding_id}}; $i++ ) {
        my $id         = $values->{_tcod_content_encoding_id}[$i];
        my $layer_type = $values->{_tcod_content_encoding_layer_type}[$i];
        my $layer_id;

        if( exists $values->{_tcod_content_encoding_layer_id} ) {
            $layer_id = $values->{_tcod_content_encoding_layer_id}[$i];
        }

        if( exists $encodings{$id} && !defined $layer_id ) {
            error( $0, $filename, $dataname,
                   "content encoding '$id' has more than unnumbered layer",
                   'cannot unambiguously reconstruct encoding stack' );
            die;
        }

        $layer_id = 0 if !defined $layer_id;
        if( int($layer_id) != $layer_id ) {
            error( $0, $filename, $dataname, 'the detected content encoding '
                 . "layer ID '$layer_id' is not an integer", undef );
            die;
        }

        if( !exists $encodings{$id} ) {
            $encodings{$id} = {};
        }

        if( !exists $encodings{$id}{$layer_id} ) {
            $encodings{$id}{$layer_id} = $layer_type;
        } else {
            error( $0, $filename, $dataname,
                   "more than one content encoding layer numbered " .
                   "'$layer_id' detected", undef );
            die;
        }
    }

    my %encodings_now;
    for my $stack (keys %encodings) {
        $encodings_now{$stack} = [ map { $encodings{$stack}{$_} }
                                   sort keys %{$encodings{$stack}} ];
    }
    return \%encodings_now;
}

#===============================================================#
# @COD::Spacegroups::Lookup::COD::table =
# (
# {
#     number          => 1,
#     hall            => ' P 1',
#     schoenflies     => 'C1^1',
#     hermann_mauguin => 'P 1',
#     universal_h_m   => 'P 1',
#     symops => [
#         'x,y,z',
#     ],
#     ncsym => [
#         'x,y,z',
#     ]
# },
# );

sub lookup_symops
{
    my ($option, $param) = @_;

    $param =~ s/ //g;
    $param =~ s/_//g;

    my $sg_full = $sg_name_abbrev{$param};

    $sg_full = "" unless defined $sg_full;
    $sg_full =~ s/\s+//g;

    foreach my $hash (@COD::Spacegroups::Lookup::COD::table,
                      @COD::Spacegroups::Lookup::COD::extra_settings)
    {
        my $value = $hash->{$option};
        $value =~ s/ //g;
        $value =~ s/_//g;

        if( $value eq $param || $value eq $sg_full )
        {
            return $hash->{symops};
        }
    }
    return undef;
}

1;
